#ifndef f_INPUTFILEMPEG2_H
#define f_INPUTFILEMPEG2_H

#include "vdinputdriver.h"
#include "vdunknown.h"
#include "infile.h"
#include "append.h"

#define MAX_AUDIO_STREAMS 8

// Do we expose multiple audio streams to the host?
#define MULTIAUDIO


// Audio types we support
#define AUDIOTYPE_AC3	0x80
#define AUDIOTYPE_LPCM	0xA0
#define AUDIOTYPE_MPEG	0xC0
#define AUDIOTYPE_MASK	0xE0

// Frame types
#define MPEG_FRAME_TYPE_I	1
#define MPEG_FRAME_TYPE_P	2
#define MPEG_FRAME_TYPE_B	3
#define MPEG_FRAME_TYPE_D	4

// Flags for MPEGSampleInfo.bitflags
#define PICTUREFLAG_RFF		0x04	// repeat_first_field
#define PICTUREFLAG_TFF		0x08	// top_field_first
#define PICTUREFLAG_PF		0x10	// progressive_frame

// Picture structures
#define TOP_FIELD		1
#define BOTTOM_FIELD	2
#define FRAME_PICTURE	3

//////////////////////////////////////////////////////////////////////////
//
//
//					MPEGPacketInfo/MPEGSampleInfo
//
//
//////////////////////////////////////////////////////////////////////////

// For MPEGPacketInfo, the top 48 bits is the unsigned file offset
// of the packet, and the bottom 16 bits is the unsigned size of
// the packet.  Thus the MPEG cannot be larger than 256 TiB - 1.
// In actual practice, we'll probably exceed the maximum number
// of packets (2 GiB - 1) well before we hit the file size limit.

typedef struct {
	unsigned __int64	file_pos;
} MPEGPacketInfo;


// MPEGSampleInfo (12 bytes)

typedef struct {
	long			size;			// size of the sample in bytes
	unsigned long	pktindex;		// packet where the sample begins
	unsigned short	pktoffset;		// offset of sample in the packet
	union {
		struct {
			unsigned char	frame_type;	// I,P,B,D
			unsigned char	bitflags;	// rff/tff/pf/struc
		};
		struct {
			unsigned short	header;		// compacted audio header
		};
	};
} MPEGSampleInfo;


class MPEGFileParser;

class InputFileMPEG2 : public vdxunknown<IVDXInputFile> {

friend MPEGFileParser;

private:
	bool fInterleaved;
	bool fUserAborted;

	// Video
	MPEGPacketInfo *video_packet_list;
	unsigned long vpackets;
	__int64 vfirstPTS;
	bool progressive_sequence;
	long lFirstCode;

	// Audio
	unsigned char audio_stream_list[96];	// 0x80 to 0xDF
	MPEGPacketInfo *audio_packet_list[48];
	unsigned long apackets[48];
	__int64 afirstPTS[48];
//	unsigned char frame_buffer[4096];
//	unsigned long prev_sample;
//	unsigned long cur_sample;
//	__int64 cur_pos;

#ifndef MULTIAUDIO
	static INT_PTR CALLBACK ChooseAudioDlg(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
	int iChosenAudio;
#endif

	inFile mFile;
	int mNumFiles;

	bool PostInit();

	// File Information dialog
	static void UpdateDlgAudio(HWND hDlg, unsigned int id);
	static INT_PTR CALLBACK InfoDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

	// Indexing
	LPWSTR GetIndexName(LPCWSTR szFile);
	bool ReadIndex(LPCWSTR szFile, AppendNames *an);
	void WriteIndex(LPCWSTR szFile, AppendNames *an);
	void WriteSampleInfo(inFile& idxFile, const MPEGSampleInfo *msi, int samples);
	void WritePacketInfo(inFile& idxFile, const MPEGPacketInfo *mpi, unsigned long packets);

	// Version 2.0 index
	bool ReadIndex20(inFile& idxFile, AppendNames *an);
	void ReadSampleInfo19(inFile& idxFile, MPEGSampleInfo *msi, int samples);
	void ReadPacketInfo19(inFile& idxFile, MPEGPacketInfo *mpi, unsigned long packets);

public:
	InputFileMPEG2(const VDXInputDriverContext& context);
	~InputFileMPEG2();

	void VDXAPIENTRY Init(const wchar_t *szFile, IVDXInputOptions *opts);
	bool VDXAPIENTRY Append(const wchar_t *szFile);

	bool VDXAPIENTRY PromptForOptions(VDXHWND, IVDXInputOptions **);
	bool VDXAPIENTRY CreateOptions(const void *buf, uint32 len, IVDXInputOptions **);
	void VDXAPIENTRY DisplayInfo(VDXHWND hwndParent);

	bool VDXAPIENTRY GetVideoSource(int index, IVDXVideoSource **);
	bool VDXAPIENTRY GetAudioSource(int index, IVDXAudioSource **);

	const VDXInputDriverContext& mContext;

	// Audio samples and audio stream info
	MPEGSampleInfo *audio_sample_list[48];
	int aframes[48];
	bool fAllowDSC;

	// Video samples and video stream info
	MPEGSampleInfo *video_sample_list;
	int vframes;
	long largest_framesize;
	int width, height;
	VDXFraction mFrameRate;
	int aspect_ratio;	// aspect_ratio_information
	long seq_ext;
	long *video_field_map;
	long fields;
	bool fAllowMatrix;
	bool fIgnoreSeqHdrs;

//	void ReadAudio(void *buffer, __int64 pos, long len, int stream_id);
	void ReadSample(void *buffer, long sample, long len, int stream_id);
};

#endif	// f_INPUTFILEMPEG2_H