///////////////////////////////////////////////////////////////////////
// MPEG-2 Plugin for VirtualDub 1.8.1+
// Traduction franaise de Sparad0x
// Copyright (C) 2008 Sparad0x
// Copyright (C) 2007-2008 fccHandler
// Copyright (C) 1998-2008 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// Main Plugin interfaces

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <math.h>

#include "infile.h"
#include "append.h"
#include "decode.h"
#include "video.h"
#include "audio.h"
#include "parser.h"

#include "resource.h"

#include <crtdbg.h>

///////////////////////////////////////////////////////////////////////////////


HMODULE g_hModule;

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved) {
	if (fdwReason == DLL_PROCESS_ATTACH) {
		g_hModule = (HMODULE)hinstDLL;
	}
	return TRUE;
}


//////////////////////////////////////////////////////////////////////////
//
//
//					InputOptionsMPEG2
//
//
//////////////////////////////////////////////////////////////////////////

class InputOptionsMPEG2 : public vdxunknown<IVDXInputOptions> {
private:
	static INT_PTR CALLBACK OptionsDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);

public:
	InputOptionsMPEG2();
	~InputOptionsMPEG2();

	uint32 VDXAPIENTRY Write(void *buf, uint32 buflen);
	void OptionsDlg(HWND hwndParent);

	struct OptionsMPEG2 {
		uint32 len;
		bool fMultipleOpen;
		bool fCreateIndex;
		bool fAllowDSC;
		bool fAllowMatrix;
//		bool fIgnoreSeqHdrs;
	} opts;
};

InputOptionsMPEG2::InputOptionsMPEG2() {
	opts.len = sizeof(opts);
	opts.fMultipleOpen = false;
	opts.fCreateIndex = false;
	opts.fAllowDSC = false;
	opts.fAllowMatrix = false;
//	opts.fIgnoreSeqHdrs = false;
}

InputOptionsMPEG2::~InputOptionsMPEG2() {
}

uint32 InputOptionsMPEG2::Write(void *buf, uint32 buflen) {
	if (buf != NULL && buflen >= opts.len) {
		memcpy(buf, &opts, opts.len);
	}
	return opts.len;
}

INT_PTR CALLBACK InputOptionsMPEG2::OptionsDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam) {
	InputOptionsMPEG2 *thisPtr;

	switch (message) {

	case WM_INITDIALOG:
		SetWindowLongPtr(hDlg, DWLP_USER, lParam);
		thisPtr = (InputOptionsMPEG2 *)lParam;
		CheckDlgButton(hDlg, IDC_MULTIPLE, thisPtr->opts.fMultipleOpen? BST_CHECKED: BST_UNCHECKED);
		CheckDlgButton(hDlg, IDC_WRITE_INDEX, thisPtr->opts.fCreateIndex? BST_CHECKED: BST_UNCHECKED);
		CheckDlgButton(hDlg, IDC_ALLOW_DSC, thisPtr->opts.fAllowDSC? BST_CHECKED: BST_UNCHECKED);
		CheckDlgButton(hDlg, IDC_ALLOW_MATRIX, thisPtr->opts.fAllowMatrix? BST_CHECKED: BST_UNCHECKED);
//		CheckDlgButton(hDlg, IDC_IGNORE_SEQHDRS, thisPtr->opts.fIgnoreSeqHdrs? BST_CHECKED: BST_UNCHECKED);
		return TRUE;

	case WM_COMMAND:
		if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) {
			thisPtr = (InputOptionsMPEG2 *)GetWindowLongPtr(hDlg, DWLP_USER);
			if (LOWORD(wParam) == IDOK) {
				thisPtr->opts.fMultipleOpen = (IsDlgButtonChecked(hDlg, IDC_MULTIPLE) == BST_CHECKED);
				thisPtr->opts.fCreateIndex = (IsDlgButtonChecked(hDlg, IDC_WRITE_INDEX) == BST_CHECKED);
				thisPtr->opts.fAllowDSC = (IsDlgButtonChecked(hDlg, IDC_ALLOW_DSC) == BST_CHECKED);
				thisPtr->opts.fAllowMatrix = (IsDlgButtonChecked(hDlg, IDC_ALLOW_MATRIX) == BST_CHECKED);
//				thisPtr->opts.fIgnoreSeqHdrs = (IsDlgButtonChecked(hDlg, IDC_IGNORE_SEQHDRS) == BST_CHECKED);
			}
			EndDialog(hDlg, TRUE);
			return TRUE;
		}

	default:
		break;
	}

	return FALSE;
}

void InputOptionsMPEG2::OptionsDlg(HWND hwndParent) {
	DialogBoxParam(g_hModule, MAKEINTRESOURCE(IDD_OPTIONS),
		hwndParent, OptionsDlgProc, (LPARAM)this);
}


///////////////////////////////////////////////////////////////////////////////
//
//
//					InputFileMPEG2
//
//
//////////////////////////////////////////////////////////////////////////

InputFileMPEG2::InputFileMPEG2(const VDXInputDriverContext& context)
	: mContext(context)
{
_RPT0(_CRT_WARN, "InputFileMPEG2 constructed\n");

	// clear variables
	mNumFiles = 0;

	// video
	video_packet_list = NULL;
	video_sample_list = NULL;
	video_field_map = NULL;
	vfirstPTS = 0;
	vpackets = 0;
	vframes = 0;
	fields = 0;
	seq_ext = 0;
	fAllowMatrix = false;
//	fIgnoreSeqHdrs = false;

	// audio
	for (int i = 0; i < 48; ++i) {
		audio_packet_list[i] = NULL;
		audio_sample_list[i] = NULL;
		afirstPTS[i] = 0;
		apackets[i] = 0;
		aframes[i] = 0;
	}
	fAllowDSC = false;
	memset(audio_stream_list, 0, 96);	
//	prev_sample = -1;
//	cur_sample = 0;
//	cur_pos = 0;

#ifndef MULTIAUDIO
	iChosenAudio = -1;
#endif
}

InputFileMPEG2::~InputFileMPEG2() {
_RPT0(_CRT_WARN, "InputFileMPEG2 destructed\n");
	delete [] video_field_map;
    delete [] video_sample_list;
	for (int i = 47; i >= 0; --i) {
		delete [] audio_sample_list[i];
		delete [] audio_packet_list[i];
	}
    delete [] video_packet_list;
}

#ifndef MULTIAUDIO


// Create MPEG audio dialog in memory

//IDD_CHOOSEAUDIO DIALOG DISCARDABLE  0, 0, 256, 96
//STYLE DS_SETFOREGROUND | WS_VISIBLE | WS_CAPTION
//CAPTION "Multiple Audio Streams Found"
//FONT 8, "MS Sans Serif"
//BEGIN
//    CTEXT           "Choose an audio stream:",IDC_STATIC,87,4,82,8
//    LISTBOX         IDC_LIST1,4,16,248,56,LBS_NOINTEGRALHEIGHT | WS_VSCROLL | 
//                    WS_TABSTOP
//    DEFPUSHBUTTON   "OK",IDOK,100,76,56,16
//END

// DS_SETFOREGROUND = 0x00000200
// WS_VISIBLE       = 0x10000000
// WS_CAPTION       = 0x00C00000

const unsigned short IDD_CHOOSEAUDIO2[] = {
	// DLGTEMPLATE
	0x0240, 0x10C0,		// style (DS_SETFONT | DS_SETFOREGROUND | WS_VISIBLE | WS_CAPTION)
	0, 0,				// dwExtendedStyle
	3,					// cdit
	0, 0, 256, 96,		// x, y, cx, cy
	0, 0,				// menu, class
	'M','u','l','t','i','p','l','e',' ','A','u','d','i','o',' ',
	'S','t','r','e','a','m','s',' ','F','o','u','n','d',' ',0,
	8,					// pointsize
	'M','S',' ','S','a','n','s',' ','S','e','r','i','f',0,

	// DLGITEMTEMPLATE
	0x0001, 0x5002,		// style (WS_CHILD | WS_VISIBLE | WS_GROUP | SS_CENTER)
	0, 0,				// dwExtendedStyle
	87, 4, 82, 8,		// x, y, cx, cy
	0xFFFF,				// id (IDC_STATIC)
	0xFFFF, 0x0082,		// class (Static)
	'C','h','o','o','s','e',' ','a','n',' ',
	'a','u','d','i','o',' ','s','t','r','e','a','m',':',0,
	0,					// data

	// DLGITEMTEMPLATE
	0x0101, 0x50A1,		// style (WS_CHILD | WS_VISIBLE | WS_BORDER |
	// WS_VSCROLL | WS_TABSTOP | LBS_NOINTEGRALHEIGHT | LBS_NOTIFY)
	0, 0,				// dwExtendedStyle
	4, 16, 248, 56,		// x, y, cx, cy
	IDC_LIST1,			// id
	0xFFFF, 0x0083,		// class (List Box)
	'L','i','s','t','1',0,
	0,					// data

	// DLGITEMTEMPLATE
	0x0001, 0x5001,		// style (WS_CHILD | WS_VISIBLE | WS_TABSTOP | BS_DEFPUSHBUTTON)
	0, 0,				// dwExtendedStyle
	100, 76, 56, 16,	// x, y, cx, cy
	IDOK,				// id
	0xFFFF, 0x0080,		// class (Button)
	'O','K',0,
	0					// data
};

INT_PTR CALLBACK InputFileMPEG2::ChooseAudioDlg(HWND hDlg, UINT uMsg,
	WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {

	case WM_INITDIALOG:
		{
			char buf[48];

			SetWindowLongPtr(hDlg, DWLP_USER, lParam);
			InputFileMPEG2 *thisPtr = (InputFileMPEG2 *)lParam;

			HWND hwndList = GetDlgItem(hDlg, IDC_LIST1);
			int i;

			// Any MPEG audio streams (C0 to DF)
			for (i = 64; i < 96; ++i) {
				if (thisPtr->audio_stream_list[i]) {
					sprintf(buf, "Flux Audio MPEG 0x%02X", i + 128);
					SendMessage(hwndList, LB_ADDSTRING, 0, (LPARAM)buf);
				}
			}

			// Any AC-3 audio streams?
			for (i = 0; i < 8; ++i) {
				if (thisPtr->audio_stream_list[i]) {
					sprintf(buf, "Audio Dolby AC-3 (substream 0x%02X)", i + 128);
					SendMessage(hwndList, LB_ADDSTRING, 0, (LPARAM)buf);
				}
			}

			// Any LPCM audio streams?
			for (i = 32; i < 40; ++i) {
				if (thisPtr->audio_stream_list[i]) {
					sprintf(buf, "Audio PCM Linaire (substream 0x%02X)", i + 128);
					SendMessage(hwndList, LB_ADDSTRING, 0, (LPARAM)buf);
				}
			}

			SendMessage(hwndList,LB_SETCURSEL,0,0);
			return TRUE;
		}

	case WM_COMMAND:
		if (LOWORD(wParam) == IDOK) {
			int i, sel;

			InputFileMPEG2 *thisPtr = (InputFileMPEG2 *)GetWindowLongPtr(hDlg, DWLP_USER);

			HWND hwndList = GetDlgItem(hDlg, IDC_LIST1);
			sel = SendMessage(hwndList, LB_GETCURSEL, 0, 0);

			// Any MPEG audio streams (C0 to DF)
			for (i = 64; i < 96; ++i) {
				if (thisPtr->audio_stream_list[i]) {
					if (sel == 0) break;
					--sel;
				}
			}

			if (i == 96) {
				// Any AC-3 audio streams?
				for (i = 0; i < 8; ++i) {
					if (thisPtr->audio_stream_list[i]) {
						if (sel == 0) break;
						--sel;
					}
				}

				if (i == 8) {
					// Any LPCM audio streams?
					for (i = 32; i < 40; ++i) {
						if (thisPtr->audio_stream_list[i]) {
							if (sel == 0) break;
							--sel;
						}
					}
				}
			}

			EndDialog(hDlg, i);
			return TRUE;
		}
		break;
	}

	return FALSE;
}

#endif	// MULTIAUDIO


void InputFileMPEG2::Init(const wchar_t *szFile, IVDXInputOptions *opts) {
	AppendNames an;
	bool fHasIndex;
	int i;

	// input options (if any) already exist at this point
	const InputOptionsMPEG2 *pOptions = (InputOptionsMPEG2 *)opts;

	if (pOptions != NULL) {
		fAllowDSC = pOptions->opts.fAllowDSC;
		fAllowMatrix = pOptions->opts.fAllowMatrix;
//		fIgnoreSeqHdrs = pOptions->opts.fIgnoreSeqHdrs;
	} else {
		fAllowDSC = false;
		fAllowMatrix = false;
//		fIgnoreSeqHdrs = false;
	}

	fHasIndex = ReadIndex(szFile, &an);

	if (!fHasIndex) {
		// see if we can open the file

		i = lstrlenW(szFile);
		if (i <= 0 || i >= MAX_PATH) {
			mContext.mpCallbacks->SetError("Nom de fichier erron %ls", szFile);
			return;
		}
			
		an.fName[0] = (LPWSTR)malloc((i + 1) * 2);
		if (an.fName[0] == NULL) {
			mContext.mpCallbacks->SetErrorOutOfMemory();
			return;
		}
			
		memcpy(an.fName[0], szFile, (i + 1) * 2);
		an.num_files = 1;

		bool bMultiple = (pOptions != NULL);
		if (bMultiple) bMultiple = pOptions->opts.fMultipleOpen;

		if (bMultiple) {
			AppendDlg(&an, &mFile, g_hModule, GetActiveWindow());
			if (an.num_files <= 0) {
				mContext.mpCallbacks->SetError("L'ouverture de plusieurs fichier a chou");
				return;
			}

		} else {
			if (!mFile.inOpen(an.fName[0])) {
				mContext.mpCallbacks->SetError("Impossible d'ouvrir %ls", szFile);
				return;
			}
		}

		// Create and invoke MPEGFileParser
		IMPEGFileParser *fileParser = CreateMPEGFileParser(this);

		if (fileParser != NULL) {
			fileParser->Parse();

			// Voila; the MPEG is parsed.
			fileParser->Destruct();

		} else {
			mContext.mpCallbacks->SetErrorOutOfMemory();
			return;
		}

		if ((fInterleaved && vpackets == 0) || vframes <= 0) {
			mContext.mpCallbacks->SetError("Aucun flux vido!");
			return;
		}

		// MPEG-2 aspect ratio?
		if (seq_ext) aspect_ratio += 16;
	}

	mNumFiles = an.num_files;

	if (pOptions != NULL && !fUserAborted) {
		if (pOptions->opts.fCreateIndex) WriteIndex(szFile, &an);
	}

#ifndef MULTIAUDIO
	for (i = 0; i < 96; ++i) {
		if (audio_stream_list[i]) {
			if (iChosenAudio >= 0) {
				// Multiple audio streams found
				iChosenAudio = DialogBoxIndirectParam(g_hModule,
					(LPCDLGTEMPLATE)IDD_CHOOSEAUDIO2, GetActiveWindow(),
					ChooseAudioDlg, (LPARAM)this);

				break;
			}
			iChosenAudio = i;
		}
	}
#endif	// MULTIAUDIO

	if (!PostInit()) {
		mContext.mpCallbacks->SetError("PostInit chou.");
		return;
	}
}

bool InputFileMPEG2::PostInit() {
    long i, j, count;
	long largest;
	int id;

	// If there are field pictures, they must come in pairs.
	// Combine matching pairs into frame pictures and adjust
	// the total number of frames.

	// This must be done BEFORE reordering pictures

	i = 0;
	while (i < (vframes - 1)) {
		MPEGSampleInfo *m1 = &(video_sample_list[i]);
		MPEGSampleInfo *m2 = &(video_sample_list[i+1]);

		if (((m1->bitflags & 3) == TOP_FIELD &&
			(m2->bitflags & 3) == BOTTOM_FIELD) ||
			((m1->bitflags & 3) == BOTTOM_FIELD &&
			(m2->bitflags & 3) == TOP_FIELD)) {

			if (m1->frame_type == m2->frame_type ||
				(m1->frame_type == MPEG_FRAME_TYPE_I &&
				m2->frame_type == MPEG_FRAME_TYPE_P)) {

				// Combine them, and remove the second field
				m1->size += m2->size;

				for (j = i + 1; j < (vframes - 1); ++j) {
					video_sample_list[j] = video_sample_list[j+1];
				}
				--vframes;
			}
		}
		++i;
	}


    // Now bubble-sort the frames into display order,
	for (i = 1; i < vframes; ++i) {
		if (video_sample_list[i].frame_type == MPEG_FRAME_TYPE_B) {
			MPEGSampleInfo m = video_sample_list[i];
			video_sample_list[i] = video_sample_list[i-1];
			video_sample_list[i-1] = m;
		}
	}


	// To be fully decodable, the resulting sequence must
	// begin with an I-frame, and mustn't end with a B-frame

	// Remove leading P/B frames
	for (i = 0; i < vframes; ++i) {
		if (video_sample_list[i].frame_type == MPEG_FRAME_TYPE_I) {
			break;
		}
	}
	// "i" is now a count of how many frames precede the first I-frame
	if (i > 0) {
		if (i >= vframes) return false;	// No I-frames at all?
		vframes -= i;
		for (j = 0; j < vframes; ++j) {
			video_sample_list[j] = video_sample_list[j + i];
		}
	}

	// Remove trailing B frames
	while (video_sample_list[vframes - 1].frame_type == MPEG_FRAME_TYPE_B) {
		--vframes;
	}
	if (vframes < 1) return false;

	for (id = 0; id < 48; ++id) {
		if (aframes[id] > 0 && audio_sample_list[id] != NULL) {
			// The skew is in 90 KHz ticks:
			long srate;
			long lTicksPerFrame = 0;

			long skew = (long)(afirstPTS[id] - vfirstPTS);

			// Calculate the duration of a frame in 90 KHz ticks

			if (id < 32) {
				MPEGAudioCompacted hdr(audio_sample_list[id][0].header);
				srate = (long)hdr.GetSamplingRateHz();
				if (hdr.GetLayer() == 1) {
					lTicksPerFrame = (long)(384.0 * 90000.0 / (double)srate + 0.5);
				} else {
					lTicksPerFrame = (long)(1152.0 * 90000.0 / (double)srate + 0.5);
				}
			} else if (id < 40) {
				if (ac3_compact_info(audio_sample_list[id][0].header, &srate, NULL, NULL)) {
					lTicksPerFrame = (long)(1536.0 * 90000.0 / (double)srate + 0.5);
				}
			} else {
				srate = (audio_sample_list[id][0].header & 0x30)? 96000: 48000;
				lTicksPerFrame = (long)(1536.0 * 90000.0 / (double)srate + 0.5);
			}

			if (lTicksPerFrame > 0) {
				MPEGSampleInfo *asl = audio_sample_list[id];

				if (skew < 0) {
					// Remove audio frames if the skew is negative
					skew = (-skew + 45) / lTicksPerFrame;

					// "skew" is now the number of frames to remove
					if (skew > 0) {
						for (i = skew; i < aframes[id]; ++i) {
							asl[i - skew] = asl[i];
						}
						aframes[id] -= skew;
					}

				} else if (skew > 0) {
					// Duplicate audio frames if the skew is positive
					skew = (skew + 45) / lTicksPerFrame;

					// "skew" is now the number of frames to duplicate
					if (skew > 0 && skew < aframes[id]) {
						for (i = aframes[id] - 1; i >= skew; --i) {
							asl[i] = asl[i - skew];
						}
						while (i >= 0) {
							asl[i] = asl[skew];
							--i;
						}
					}
				}
			}
		}
	}

	// Generate field map
	// While we're at it, determine the largest frame

	// "Force FILM" is a work in progress.  For now,
	// just emulate the 1.5.10 behavior and ignore RFF

	count = 0;
	j = 0;	// cadence

	for (i = 0; i < vframes; ++i) {
		const MPEGSampleInfo *m1 = &(video_sample_list[i]);

		// this is overly complicated
		if (	(m1->bitflags & PICTUREFLAG_RFF)
			&&	((m1->bitflags & 3) == FRAME_PICTURE)
			&&	(m1->bitflags & PICTUREFLAG_PF)) {

			if (this->progressive_sequence) {
				if (m1->bitflags & PICTUREFLAG_TFF) {
					count += 3;	// 3 frames
				} else {
					count += 2;	// 2 frames
				}
			} else {
				++count;		// 3 fields
				j = 1 - j;		// flip the cadence
				if (j == 0) ++count;
			}

		} else {
			++count;			// 1 frame, or 2 fields
		}
	}

	if (!(video_field_map = new long[count]))
		return false;

	// Now load the map
	fields = count;
	count = 0;

	j = 0;	// cadence
	largest = 0;

	for (i = 0; i < vframes; ++i) {
		const MPEGSampleInfo *m1 = &(video_sample_list[i]);

		if (m1->size > largest) largest = m1->size;

		if (	(m1->bitflags & PICTUREFLAG_RFF)
			&&	((m1->bitflags & 3) == FRAME_PICTURE)
			&&	(m1->bitflags & PICTUREFLAG_PF)) {

			// Add 1 progressive frame
			video_field_map[count++] = i * 2 + j;

			if (this->progressive_sequence) {
				// 2 progressive frames
				video_field_map[count++] = i * 2 + j;
				if (m1->bitflags & PICTUREFLAG_TFF) {
					// 3 progressive frames
					video_field_map[count++] = i * 2 + j;
				}
			} else {
				j = 1 - j;	// flip cadence
				if (j == 0) {
					// Add a duplicate frame
					video_field_map[count++] = i * 2 + j;
				}
			}

		} else {
			// 1 progressive frame, or 2 field pictures
			video_field_map[count++] = i * 2 + j;
		}
	}

    // Set video_packet_buffer to size of the largest frame
    count = (largest + 31) & -15;

	largest_framesize = count;
    return true;
}

bool InputFileMPEG2::Append(const wchar_t *szFile) {
	return false;
}

bool InputFileMPEG2::PromptForOptions(VDXHWND hwndParent, IVDXInputOptions **ppOptions) {
	InputOptionsMPEG2 *p = new InputOptionsMPEG2;

	*ppOptions = p;
	if (p == NULL) return false;

	p->AddRef();

	p->OptionsDlg((HWND)hwndParent);
	return true;
}

bool InputFileMPEG2::CreateOptions(const void *buf, uint32 len, IVDXInputOptions **ppOptions) {
	InputOptionsMPEG2 *p = new InputOptionsMPEG2;

	*ppOptions = p;
	if (p == NULL) return false;

	p->AddRef();

	if (len > p->opts.len) len = p->opts.len;
	memcpy(&(p->opts), buf, len);

	return true;
}

static char *time_to_string(double seconds) {

	// There are 86400 seconds in a day,
	// about 32 million seconds in a year.
	// Since we are unlikely to encounter videos
	// of such an extreme duration, we can easily
	// get away with using longs for this math...

	char *ret = (char *)malloc(32);
	if (ret != NULL) {
		long hours, mins;
		long secs = (long)seconds;

		hours = secs / 3600;
		secs = secs % 3600;
		mins = secs / 60;
		secs = secs % 60;

		if (hours > 0) {
			sprintf(ret, "%ld:%02ld:%02ld", hours, mins, secs);
		} else {
			sprintf(ret, "%ld:%02ld", mins, secs);
		}
	}
	return ret;
}


static struct {
	__int64 i64VideoSize;
	__int64 i64AudioSize;
	long lFrameCnt[3];
	long lFrameMinSize[3];
	long lFrameMaxSize[3];
	long lFrameTotalSize[3];
} pInfo;

static const char *szModes[6] = {
	"mono", "stro", "3 canaux",
	"4 canaux", "5 canaux", "5.1"
};

static const char *szLayers[4] = {
	"?", "I", "II", "III"
};


void InputFileMPEG2::UpdateDlgAudio(HWND hDlg, unsigned int index) {
	char buf[64];
	int i, id, count;
	unsigned short header;

	InputFileMPEG2 *thisPtr =
		(InputFileMPEG2 *)GetWindowLongPtr(hDlg, DWLP_USER);

	if (index >= (unsigned int)96) return;

	count = index;
	for (i = 0; i < 96; ++i) {
		id = thisPtr->audio_stream_list[i];
		if (id > 0) {
			if (count == 0) break;
			--count;
		}
	}

	if (id <= 0 || id > 48) return;
	--id;

	if (thisPtr->aframes[id] <= 0
		|| thisPtr->audio_sample_list[id] == NULL) return;

	count = 0;
	for (i = 0; i < 96; ++i) {
		if (thisPtr->audio_stream_list[i]) ++count;
	}

	sprintf(buf, "Flux %i sur %i", index + 1, count);
	SetDlgItemText(hDlg, IDC_STREAMNUM, buf);

	pInfo.i64AudioSize = 0;
	for (i = 0; i < thisPtr->aframes[id]; ++i) {
		pInfo.i64AudioSize += thisPtr->audio_sample_list[id][i].size;
	}

	header = thisPtr->audio_sample_list[id][0].header;

	if (id < 32) {
		MPEGAudioCompacted hdr(header);
		__int64 iTotalBitrate = 0;

		for (i = 0; i < thisPtr->aframes[id]; ++i) {
			iTotalBitrate += MPEGAudioCompacted(
				thisPtr->audio_sample_list[id][i].header).GetBitrateKbps();
		}

		iTotalBitrate = (__int64)
			((double)iTotalBitrate / (double)thisPtr->aframes[id] + 0.5);

        sprintf(buf, "%ldKHz %s, %ldKbps Layer %s",
			hdr.GetSamplingRateHz() / 1000,
			szModes[hdr.IsStereo()? 1: 0],
			(long)iTotalBitrate,
			szLayers[hdr.GetLayer()]);

		SetDlgItemText(hDlg, IDC_AUDIO_FORMAT, buf);

	} else if (id < 40) {
		int brate, chans;
		long srate;

		if (ac3_compact_info(header, &srate, &brate, &chans)) {
			sprintf(buf, "%ldKHz %s, %ldKbps AC-3",
				srate / 1000, szModes[chans - 1], brate);

			SetDlgItemText(hDlg, IDC_AUDIO_FORMAT, buf);
		}

	} else {
		long srate = (header & 0x30)? 96000: 48000;
		int chans = (header & 7)? 2: 1;

		sprintf(buf, "%ldKHz %s, %ldKbps LPCM",
			srate / 1000, szModes[chans - 1],
			(long)(((double)srate * (double)chans * 2.0) / 125.0 + 0.5));

		SetDlgItemText(hDlg, IDC_AUDIO_FORMAT, buf);
    }

	sprintf(buf, "%ld", thisPtr->aframes[id]);
	SetDlgItemText(hDlg, IDC_AUDIO_NUMFRAMES, buf);

	sprintf(buf, "%ldK", ((long)pInfo.i64AudioSize + 512) / 1024);
	SetDlgItemText(hDlg, IDC_AUDIO_SIZE, buf);

	// Calculate internal skew with proper rounding
	i = (thisPtr->afirstPTS[id] < thisPtr->vfirstPTS)? -1: 1;
	i = (int)(((thisPtr->afirstPTS[id] - thisPtr->vfirstPTS) + 45 * i) / 90);
	sprintf(buf, "%i ms", i);
	SetDlgItemText(hDlg, IDC_AUDIO_SKEW, buf);
}

INT_PTR CALLBACK InputFileMPEG2::InfoDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam) {
	InputFileMPEG2 *thisPtr;
	MPEGSampleInfo *msi;
	char buf[128];
	int i, count;
	unsigned int id;

	static const char *sz_aspect_ratios[32] = {
		// MPEG-1 SAR
		"SAR = 0 (interdit)",
		"SAR = 1 (1.0) 1:1 VGA",
		"SAR = 2 (0.6735)",
		"SAR = 3 (0.7031) 16:9 PAL",
		"SAR = 4 (0.7615)",
		"SAR = 5 (0.8055)",
		"SAR = 6 (0.8437) 16:9 NTSC",
		"SAR = 7 (0.8935)",
		"SAR = 8 (0.9157) 4:3 PAL",
		"SAR = 9 (0.9815)",
		"SAR = 10 (1.0255)",
		"SAR = 11 (1.0695)",
		"SAR = 12 (1.095) 4:3 NTSC",
		"SAR = 13 (1.1575)",
		"SAR = 14 (1.2015)",
		"SAR = 15 (rserv)",
		// MPEG-2 SAR/DAR
		"DAR = 0 (interdit)",
		"SAR = 1 (1.0)",
		"DAR = 2 (4:3)",
		"DAR = 3 (16:9)",
		"DAR = 4 (2.21:1)",
		"DAR = 5 (rserv)",
		"DAR = 6 (rserv)",
		"DAR = 7 (rserv)",
		"DAR = 8 (rserv)",
		"DAR = 9 (rserv)",
		"DAR = 10 (rserv)",
		"DAR = 11 (rserv)",
		"DAR = 12 (rserv)",
		"DAR = 13 (rserv)",
		"DAR = 14 (rserv)",
		"DAR = 15 (rserv)"
	};

    static const UINT uiCtlIds[3] = {
		IDC_VIDEO_IFRAMES, IDC_VIDEO_PFRAMES, IDC_VIDEO_BFRAMES
	};


	switch (message) {

    case WM_INITDIALOG:
		SetWindowLongPtr(hDlg, DWLP_USER, lParam);
		thisPtr = (InputFileMPEG2 *)lParam;

		memset(&pInfo, 0, sizeof(pInfo));

		for (i = 0; i < 3; ++i) {
			pInfo.lFrameMinSize[i] = 0x7FFFFFFF;
		}

        if (thisPtr->vframes > 0) {
			char *s;
			double fps;
			int i;

			s = buf + sprintf(buf, "Information MPEG-%i", thisPtr->seq_ext? 2: 1);
			if (thisPtr->mNumFiles != 1) {
				sprintf(s, " (%i fichiers)", thisPtr->mNumFiles);
			}
			SetWindowText(hDlg, buf);

			fps = (double)thisPtr->mFrameRate.mNumerator /
				(double)thisPtr->mFrameRate.mDenominator;

			sprintf(buf, "%dx%d, %.3f fps (%ld s)",
				thisPtr->width, thisPtr->height,
				fps, (long)(1000000.0 / fps + 0.5));

			SetDlgItemText(hDlg, IDC_VIDEO_FORMAT, buf);

			SetDlgItemText(hDlg, IDC_ASPECT_RATIO,
				sz_aspect_ratios[thisPtr->aspect_ratio & 31]);

			// Duration
			s = time_to_string((double)thisPtr->fields / fps + 0.5);
			if (s != NULL) {
				sprintf(buf, "%lu (%s)", (unsigned)thisPtr->fields, s);
				free(s);
			} else {
				sprintf(buf, "%lu (?)", (unsigned)thisPtr->fields);
			}
			SetDlgItemText(hDlg, IDC_VIDEO_NUMFRAMES, buf);

			for (i = 0; i < thisPtr->vframes; ++i) {
				msi = &(thisPtr->video_sample_list[i]);
				int iFrameType = msi->frame_type;

				if (iFrameType) {
					long lSize = msi->size;
					--iFrameType;

					++pInfo.lFrameCnt[iFrameType];
					pInfo.lFrameTotalSize[iFrameType] += lSize;

					if (lSize < pInfo.lFrameMinSize[iFrameType])
						pInfo.lFrameMinSize[iFrameType] = lSize;

					if (lSize > pInfo.lFrameMaxSize[iFrameType])
						pInfo.lFrameMaxSize[iFrameType] = lSize;

					pInfo.i64VideoSize += lSize;
				}
			}

            sprintf(buf, "%ld / %ld / %ld", pInfo.lFrameCnt[0], pInfo.lFrameCnt[1], pInfo.lFrameCnt[2]);
            SetDlgItemText(hDlg, IDC_VIDEO_FRAMETYPECNT, buf);

			for (i = 0; i < 3; ++i) {

				if (pInfo.lFrameCnt[i]) {
					sprintf(buf, "%ld / %ld / %ld (%ldK)",
						pInfo.lFrameMinSize[i],
						(long)((double)pInfo.lFrameTotalSize[i] / (double)pInfo.lFrameCnt[i] + 0.5),
						pInfo.lFrameMaxSize[i],
						(pInfo.lFrameTotalSize[i] + 512) >> 10);
				} else {
					sprintf(buf,"(aucune image-%c)", "IPB"[i]);
				}

				SetDlgItemText(hDlg, uiCtlIds[i], buf);
			}


	// This won't work for movies with RFF, because we don't know
	// the display rate of the encoded frames (it may even vary).
	// However, we do know the playback duration, so divide that
	// by the number of coded frames to get an average frame rate.

	// bits * (frames per sec) / frames = bits per sec
	// (a / b) / c  =  a / (b * c)
	// a / (b / c)  =  (a * c) / b

	// First, calculate the AVERAGE frame rate:

		// calculate the duration in seconds:
			fps = (double)thisPtr->fields / fps;

		// divide the total size of the video stream by that
		// number to produce the rate in bytes per second:
			fps = (double)pInfo.i64VideoSize / fps;

			sprintf(buf, "%ld Kbps (%ldKo/s)",
				(long)(fps * 0.008 + 0.5),
				(long)(fps / 1024.0 + 0.5));

			SetDlgItemText(hDlg, IDC_VIDEO_AVGBITRATE, buf);
		}

		/////////// Audio ///////////////////////////////////////

		for (i = 0; i < 96; ++i) {
			id = thisPtr->audio_stream_list[i] - 1;
			if (id < 48U) break;
		}
		if (i == 96) return TRUE;

		count = 0;
		for (i = 0; i < 96; ++i) {
			if (thisPtr->audio_stream_list[i]) ++count;
		}

		if (count > 1) {
			HWND hwndSpin = GetDlgItem(hDlg, IDC_SPIN1);
			EnableWindow(hwndSpin, TRUE);
			SendMessage(hwndSpin, UDM_SETRANGE, 0,
				(LPARAM)MAKELONG((count - 1), 0));
			SendMessage(hwndSpin, UDM_SETPOS, 0,
				(LPARAM)MAKELONG(0, 0));
			ShowWindow(hwndSpin, SW_SHOW);
		}

		UpdateDlgAudio(hDlg, 0);
		return TRUE;


	case WM_COMMAND:
		// Note: "IDCANCEL" is the [X] button
		// Return TRUE if we process the message

		if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) {
			EndDialog(hDlg, TRUE);
			return TRUE;
		}
		break;

	case WM_NOTIFY:
		{
			const NMUPDOWN *nmud = (NMUPDOWN *)lParam;

			if (nmud->hdr.idFrom == IDC_SPIN1 &&
				nmud->hdr.code == UDN_DELTAPOS)
			{
				UpdateDlgAudio(hDlg, nmud->iPos + nmud->iDelta);
			}
		}
		break;

    }

    return FALSE;
}

void InputFileMPEG2::DisplayInfo(VDXHWND hwndParent) {
    DialogBoxParam(g_hModule, MAKEINTRESOURCE(IDD_INFORMATION),
		(HWND)hwndParent, InfoDlgProc, (LPARAM)this);
}

bool InputFileMPEG2::GetVideoSource(int index, IVDXVideoSource **ppVS) {
	*ppVS = NULL;

	if (index) return false;

	IVDXVideoSource *pVS = CreateVideoSourceMPEG2(this);
	if (!pVS) return false;

	*ppVS = pVS;
	pVS->AddRef();
	return true;
}


bool InputFileMPEG2::GetAudioSource(int index, IVDXAudioSource **ppAS) {
	unsigned int id;
	int i;

	IVDXAudioSource *pAS = NULL;
	*ppAS = NULL;

#ifdef MULTIAUDIO
	for (i = 0; i < 96; ++i) {
		id = audio_stream_list[i] - 1;
		if (id < 48U) {
			if (index == 0) break;
			--index;
		}
	}
#else
	i = iChosenAudio;
	if (index != 0 || (unsigned int)i >= 96U) return false;
	id = audio_stream_list[i] - 1;
#endif

	if (i < 96) {
		if (id < 32) {
			if (fAllowDSC) {
				pAS = CreateAudioSourceMPEG(this, id);
			} else {
				pAS = CreateAudioSourceMPEG2(this, id);
			}
		} else if (id < 40) {
			pAS = CreateAudioSourceAC3(this, id);
		} else {
			pAS = CreateAudioSourceLPCM(this, id);
		}
	}

	if (pAS == NULL) return false;

	*ppAS = pAS;
	pAS->AddRef();
	return true;
}

/*
void InputFileMPEG2::ReadAudio(void *buffer, __int64 pos, long len, int stream_id) {

	MPEGSampleInfo *msi;
	long offset;

	if (pos < cur_pos) {
		// Reset
		prev_sample = -1;
		cur_sample = 0;
		cur_pos = 0;
	}
	msi = &(audio_sample_list[stream_id][cur_sample]);

	while (pos >= (cur_pos + msi->size)) {
		if (cur_sample == (aframes[stream_id] - 1U)) break;
		cur_pos += msi->size;
		++cur_sample;
		++msi;
	}

	offset = (unsigned long)(pos - cur_pos);

	while (len > 0) {
		long remains = msi->size - offset;
		if (remains > len) remains = len;

		ReadSample(frame_buffer, cur_sample, remains, stream_id + 1);

		len -= remains;
		if (len == 0) break;

		if (cur_sample == (aframes[stream_id] - 1U)) {
			mContext.mpCallbacks->SetError("Attempt to read past end of audio");
			break;
		}

		cur_pos += msi->size;
		++cur_sample;
		++msi;

		offset = 0;
	}
}
*/

void InputFileMPEG2::ReadSample(void *buffer, long sample, long len, int stream_id) {
	unsigned long pkt, pkts;
	MPEGPacketInfo *packet_list;
	long delta;

	if (stream_id > 0) {
		--stream_id;
		pkts = apackets[stream_id];
		pkt = audio_sample_list[stream_id][sample].pktindex;
		delta = audio_sample_list[stream_id][sample].pktoffset;
		packet_list = audio_packet_list[stream_id];
	} else {
		pkts = vpackets;
		pkt = video_sample_list[sample].pktindex;
		delta = video_sample_list[sample].pktoffset;
		packet_list = video_packet_list;
	}

	while (len > 0) {
		if (pkt >= pkts) {
			if (stream_id > 0) {
				mContext.mpCallbacks->SetError("Tentative de lecture au-del de l'audio");
			} else {
				mContext.mpCallbacks->SetError("Tentative de lecture au-del de la vido");
			}
			break;
		}

		long tc = (long)(packet_list[pkt].file_pos & 0xFFFF) - delta;

		if (tc > len) tc = len;

		mFile.inSeek((packet_list[pkt].file_pos >> 16) + delta);
		mFile.inRead(buffer, tc);

		len -= tc;
		buffer = (char *)buffer + tc;
		++pkt;
		delta = 0;
	}
}


///////////////////////////////////////////////////////////////////////////////
//
//
//					InputFileDriverMPEG2
//
//
//////////////////////////////////////////////////////////////////////////

class InputFileDriverMPEG2 : public vdxunknown<IVDXInputFileDriver> {
public:
	InputFileDriverMPEG2(const VDXInputDriverContext& context);
	~InputFileDriverMPEG2();

	int VDXAPIENTRY AddRef();
	int VDXAPIENTRY Release();

	int		VDXAPIENTRY DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize);
	bool	VDXAPIENTRY CreateInputFile(uint32 flags, IVDXInputFile **ppFile);

protected:
	int		mRefCount;

	const VDXInputDriverContext& mContext;
};

InputFileDriverMPEG2::InputFileDriverMPEG2(const VDXInputDriverContext& context)
	: mRefCount(0)
	, mContext(context)
{
}

InputFileDriverMPEG2::~InputFileDriverMPEG2() {
}

int VDXAPIENTRY InputFileDriverMPEG2::AddRef() {
	return ++mRefCount;
}

int VDXAPIENTRY InputFileDriverMPEG2::Release() {
	int rv = --mRefCount;
	if (!rv) delete this;
	return rv;
}

int VDXAPIENTRY InputFileDriverMPEG2::DetectBySignature(
	const void *pHeader, sint32 nHeaderSize,
	const void *pFooter, sint32 nFooterSize, sint64 nFileSize)
{
/*
	if (nHeaderSize >= 4) {
		unsigned long hdr;

		__asm {
			mov		eax, pHeader
			mov		eax, [eax]
			bswap	eax
			mov		hdr, eax
		}

		if (hdr == 0x1B3 || hdr == 0x1BA) return 1;
		if (hdr >= 0x1C0 && hdr <= 0x1EF) return 0;
	}
*/
	return -1;
}

bool VDXAPIENTRY InputFileDriverMPEG2::CreateInputFile(uint32 flags, IVDXInputFile **ppFile) {
	IVDXInputFile *p = new InputFileMPEG2(mContext);
	if (!p) return false;

	*ppFile = p;
	p->AddRef();
	return true;
}


bool VDXAPIENTRY mpeg2_create(const VDXInputDriverContext *pContext, IVDXInputFileDriver **ppDriver) {
	IVDXInputFileDriver *p = new InputFileDriverMPEG2(*pContext);
	if (!p) return false;

	*ppDriver = p;
	p->AddRef();
	return true;
}


const uint8 MPEG2_sig[] = {
	0x00, 0xFF,
	0x00, 0xFF,
	0x01, 0xFF,
//	0xB2, 0xF6		// either 0x1B3 or 0x1BA
    0x80, 0x80
};

//	0010
//	0110
//	x01x	anything can go here
//	0010	0x1B2
//	0011	0x1B3
//	1010	0x1BA
//	1011	0x1BB

const VDXInputDriverDefinition mpeg2_input = {
	sizeof(VDXInputDriverDefinition),		// size of this structure
	VDXInputDriverDefinition::kFlagSupportsVideo |
	VDXInputDriverDefinition::kFlagSupportsAudio, // |
//	VDXInputDriverDefinition::kFlagCustomSignature,
	1,										// priority
	sizeof(MPEG2_sig),						// size of file signature
	MPEG2_sig,								// pointer to signature
	L"*.mpg|*.vob|*.m2v|*.mpeg|*.vdr",			// extension pattern
	L"Fichiers MPEG-2 (*.mpg;*.vob;*.m2v;*.mpeg;*.vdr)|*.mpg;*.vob;*.m2v;*.mpeg;*.vdr",
	L"MPEG-2",								// driver target name
	mpeg2_create							// create proc
};

const VDXPluginInfo mpeg2_info = {
	sizeof(VDXPluginInfo),				// size of this structure
	L"MPEG-2",							// name of this plugin
	L"fccHandler",						// author's name
	L"Charge des flux MPEG-2.",			// plugin description
	0x03010000,							// version = 3.1.0
	kVDXPluginType_Input,				// type of plugin = input
	0,									// flags = 0
	10,									// API version required
	kVDXPlugin_APIVersion,				// API version used
	2,									// API version required
	kVDXPlugin_InputDriverAPIVersion,	// API version used
	&mpeg2_input						// driver definition
};


const uint8 MIDX_sig[] = {
	'M', 0xFF,
	'I', 0xFF,
	'D', 0xFF,
	'X', 0xFF,
	0, 0xFF,		// version 2.0
	2, 0xFF,
	0, 0xFF,
	0, 0xFF
};

const VDXInputDriverDefinition midx_input = {
	sizeof(VDXInputDriverDefinition),		// size of this structure
	VDXInputDriverDefinition::kFlagSupportsVideo,
	0,										// priority
	sizeof(MIDX_sig),						// size of file signature
	MIDX_sig,								// pointer to signature
	L"*.midx",								// extension pattern
	L"Index MPEG-2 (*.midx)|*.midx",
	L"MIDX",								// driver target name
	mpeg2_create							// create proc
};

const VDXPluginInfo midx_info = {
	sizeof(VDXPluginInfo),					// size of this structure
	L"MIDX",								// name of this plugin
	L"fccHandler",							// author's name
	L"Charge des fichiers index MPEG-2.",	// plugin description
	0x02000000,								// version = 2.0.0
	kVDXPluginType_Input,					// type of plugin = input
	0,										// flags = 0
	10,										// API version required
	kVDXPlugin_APIVersion,					// API version used
	2,										// API version required
	kVDXPlugin_InputDriverAPIVersion,		// API version used
	&midx_input								// driver definition
};


///////////////////////////////////////////////////////////////////////////

const VDXPluginInfo *const kPlugins[] = {
	&mpeg2_info,
	&midx_info,
	NULL
};

extern "C" __declspec(dllexport) const VDXPluginInfo *const * VDXAPIENTRY VDGetPluginInfo() {
	return kPlugins;
}
